import asyncio

import pytest

from aiormq.tools import Countdown, awaitable


def simple_func():
    return 1


def await_result_func():
    return asyncio.sleep(0)


async def await_func():
    await asyncio.sleep(0)
    return 2


def return_future():
    loop = asyncio.get_event_loop()
    f = loop.create_future()
    loop.call_soon(f.set_result, 5)
    return f


async def await_future():
    return (await return_future()) + 1


def return_coroutine():
    return await_future()


AWAITABLE_FUNCS = [
    (simple_func, 1),
    (await_result_func, None),
    (await_func, 2),
    (return_future, 5),
    (await_future, 6),
    (return_coroutine, 6),
]


@pytest.mark.parametrize("func,result", AWAITABLE_FUNCS)
async def test_awaitable(func, result):
    assert await awaitable(func)() == result


async def test_countdown():
    countdown = Countdown(timeout=0.1)
    await countdown(asyncio.sleep(0))

    # waiting for the countdown exceeded
    await asyncio.sleep(0.2)

    task = asyncio.create_task(asyncio.sleep(0))

    with pytest.raises(asyncio.TimeoutError):
        await countdown(task)

    assert task.cancelled()



import asyncio
import itertools
import os
import ssl
import uuid
from binascii import hexlify
from typing import Any, Optional, Tuple

import aiomisc
import pytest
from pamqp.commands import Basic
from yarl import URL

import aiormq
from aiormq.abc import DeliveredMessage, SSLCerts
from aiormq.auth import AuthBase, ExternalAuth, PlainAuth
from aiormq.connection import (
    SSLContextProvider,
    TransportFactory,
    parse_int,
    parse_timeout,
    parse_bool
)

from .conftest import cert_path


CERT_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), "certs"))


async def test_simple(amqp_connection: aiormq.Connection):
    channel1 = await amqp_connection.channel()
    await channel1.basic_qos(prefetch_count=1)

    assert channel1.number

    channel2 = await amqp_connection.channel(11)
    assert channel2.number

    await channel1.close()
    await channel2.close()

    channel = await amqp_connection.channel()

    queue = asyncio.Queue()

    deaclare_ok = await channel.queue_declare(auto_delete=True)
    consume_ok = await channel.basic_consume(deaclare_ok.queue, queue.put)
    await channel.basic_publish(b"foo", routing_key=deaclare_ok.queue)

    message: DeliveredMessage = await queue.get()
    assert message.body == b"foo"

    with pytest.raises(aiormq.exceptions.PublishError) as e:
        await channel.basic_publish(
            b"bar", routing_key=deaclare_ok.queue + "foo", mandatory=True,
        )

    message = e.value.message

    assert message.delivery.routing_key == deaclare_ok.queue + "foo"
    assert message.body == b"bar"
    assert "'NO_ROUTE' for routing key" in repr(e.value)

    cancel_ok = await channel.basic_cancel(consume_ok.consumer_tag)
    assert cancel_ok.consumer_tag == consume_ok.consumer_tag
    await channel.queue_delete(deaclare_ok.queue)

    deaclare_ok = await channel.queue_declare(auto_delete=True)
    await channel.basic_publish(b"foo bar", routing_key=deaclare_ok.queue)

    message = await channel.basic_get(deaclare_ok.queue, no_ack=True)
    assert message.body == b"foo bar"

    await amqp_connection.close()

    with pytest.raises(RuntimeError):
        await channel.basic_get(deaclare_ok.queue)

    with pytest.raises(RuntimeError):
        await amqp_connection.channel()


async def test_channel_reuse(amqp_connection: aiormq.Connection):
    for _ in range(10):
        channel = await amqp_connection.channel(channel_number=1)
        await channel.basic_qos(prefetch_count=1)
        await channel.close()
        del channel


async def test_channel_closed_reuse(amqp_connection: aiormq.Connection):
    for _ in range(10):
        channel = await amqp_connection.channel(channel_number=1)
        with pytest.raises(aiormq.ChannelAccessRefused):
            await channel.exchange_declare("", passive=True, auto_delete=True)


async def test_bad_channel(amqp_connection: aiormq.Connection):
    with pytest.raises(ValueError):
        await amqp_connection.channel(65536)

    with pytest.raises(ValueError):
        await amqp_connection.channel(-1)

    channel = await amqp_connection.channel(65535)

    with pytest.raises(aiormq.exceptions.ChannelNotFoundEntity):
        await channel.queue_bind(uuid.uuid4().hex, uuid.uuid4().hex)


async def test_properties(amqp_connection):
    assert amqp_connection.connection_tune.channel_max > 0
    assert amqp_connection.connection_tune.heartbeat > 1
    assert amqp_connection.connection_tune.frame_max > 1024
    await amqp_connection.close()


async def test_open(amqp_connection):
    with pytest.raises(RuntimeError):
        await amqp_connection.connect()

    channel = await amqp_connection.channel()
    await channel.close()
    await amqp_connection.close()


class _TcpTransportFactory(TransportFactory):
    async def create(
            self,
            url: URL,
            **kwargs: Any,
    ) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        ssl_context_provider = kwargs.pop("ssl_context_provider")
        assert isinstance(ssl_context_provider, SSLContextProvider)

        loop = asyncio.get_event_loop()
        reader = asyncio.StreamReader(loop=loop)
        protocol = asyncio.StreamReaderProtocol(reader, loop=loop)
        if url.scheme == "amqps":
            ssl = await ssl_context_provider.get_context()
        else:
            ssl = None

        transport, _ = await loop.create_connection(
            lambda: protocol, url.host, url.port, ssl=ssl,
        )
        writer = asyncio.StreamWriter(transport, protocol, reader, loop)
        return reader, writer


async def test_open_with_transport_factory(amqp_url):
    amqp_connection = await aiormq.connect(
        amqp_url,
        transport_factory=_TcpTransportFactory(),
    )

    channel = await amqp_connection.channel()
    await channel.close()
    await amqp_connection.close()


async def test_channel_close(amqp_connection):
    channel = await amqp_connection.channel()

    assert channel.number in amqp_connection.channels

    await channel.close()

    assert channel.number not in amqp_connection.channels


async def test_conncetion_reject(event_loop):
    with pytest.raises(ConnectionError):
        await aiormq.connect(
            "amqp://guest:guest@127.0.0.1:59999/", loop=event_loop,
        )

    connection = aiormq.Connection(
        "amqp://guest:guest@127.0.0.1:59999/", loop=event_loop,
    )

    with pytest.raises(ConnectionError):
        await event_loop.create_task(connection.connect())


async def test_auth_base(amqp_connection):
    with pytest.raises(NotImplementedError):
        AuthBase(amqp_connection).marshal()


async def test_auth_plain(amqp_connection, event_loop):
    auth = PlainAuth(amqp_connection).marshal()

    assert auth == PlainAuth(amqp_connection).marshal()

    auth_parts = auth.split("\x00")
    assert auth_parts == ["", "guest", "guest"]

    connection = aiormq.Connection(
        amqp_connection.url.with_user("foo").with_password("bar"),
        loop=event_loop,
    )

    auth = PlainAuth(connection).marshal()

    auth_parts = auth.split("\x00")
    assert auth_parts == ["", "foo", "bar"]

    auth = PlainAuth(connection)
    auth.value = "boo"

    assert auth.marshal() == "boo"


async def test_auth_external(amqp_direct_url):

    url = amqp_direct_url.with_scheme("amqps")
    url.update_query(auth="external")

    connection = aiormq.Connection

    auth = ExternalAuth(connection).marshal()

    auth = ExternalAuth(connection)
    auth.value = ""

    assert auth.marshal() == ""


async def test_channel_closed(amqp_connection):

    for i in range(10):
        channel = await amqp_connection.channel()

        with pytest.raises(aiormq.exceptions.ChannelNotFoundEntity):
            await channel.basic_consume("foo", lambda x: None)

        channel = await amqp_connection.channel()

        with pytest.raises(aiormq.exceptions.ChannelNotFoundEntity):
            await channel.queue_declare(
                "foo_%s" % i, auto_delete=True, passive=True,
            )

    await amqp_connection.close()


async def test_timeout_default(amqp_direct_url, event_loop):
    connection = aiormq.Connection(amqp_direct_url, loop=event_loop)
    await connection.connect()
    assert connection.timeout == 60
    await connection.close()


async def test_timeout_1000(amqp_direct_url, event_loop):
    url = amqp_direct_url.update_query(timeout=1000)
    connection = aiormq.Connection(url, loop=event_loop)
    await connection.connect()
    assert connection.timeout
    await connection.close()


async def test_heartbeat_0(amqp_direct_url, event_loop):
    url = amqp_direct_url.update_query(heartbeat=0)
    connection = aiormq.Connection(url, loop=event_loop)
    await connection.connect()
    assert connection.connection_tune.heartbeat == 0
    await connection.close()


async def test_heartbeat_default(amqp_direct_url, event_loop):
    connection = aiormq.Connection(amqp_direct_url, loop=event_loop)
    await connection.connect()
    assert connection.connection_tune.heartbeat == 60
    await connection.close()


async def test_heartbeat_above_range(amqp_direct_url, event_loop):
    url = amqp_direct_url.update_query(heartbeat=70000)
    connection = aiormq.Connection(url, loop=event_loop)
    await connection.connect()
    assert connection.connection_tune.heartbeat == 0
    await connection.close()


async def test_heartbeat_under_range(amqp_direct_url, event_loop):
    url = amqp_direct_url.update_query(heartbeat=-1)
    connection = aiormq.Connection(url, loop=event_loop)
    await connection.connect()
    assert connection.connection_tune.heartbeat == 0
    await connection.close()


async def test_heartbeat_not_int(amqp_direct_url, event_loop):
    url = amqp_direct_url.update_query(heartbeat="None")
    connection = aiormq.Connection(url, loop=event_loop)
    await connection.connect()
    assert connection.connection_tune.heartbeat == 0
    await connection.close()


async def test_bad_credentials(amqp_url: URL):
    with pytest.raises(aiormq.exceptions.ProbableAuthenticationError):
        await aiormq.connect(amqp_url.with_password(uuid.uuid4().hex))


async def test_non_publisher_confirms(amqp_connection):
    amqp_connection.server_capabilities["publisher_confirms"] = False

    with pytest.raises(ValueError):
        await amqp_connection.channel(publisher_confirms=True)

    await amqp_connection.channel(publisher_confirms=False)


@pytest.mark.slow
async def test_no_free_channels(amqp_connection: aiormq.Connection):
    await asyncio.wait_for(
        asyncio.gather(
            *[
                amqp_connection.channel(n + 1)
                for n in range(amqp_connection.connection_tune.channel_max)
            ],
        ),
        timeout=120,
    )

    with pytest.raises(aiormq.exceptions.ConnectionNotAllowed):
        await asyncio.wait_for(amqp_connection.channel(), timeout=5)


async def test_huge_message(amqp_connection: aiormq.Connection):
    conn: aiormq.Connection = amqp_connection
    body = os.urandom(int(conn.connection_tune.frame_max * 2.5))
    channel: aiormq.Channel = await conn.channel()

    queue = asyncio.Queue()

    deaclare_ok = await channel.queue_declare(auto_delete=True)
    await channel.basic_consume(deaclare_ok.queue, queue.put)
    await channel.basic_publish(body, routing_key=deaclare_ok.queue)

    message: DeliveredMessage = await queue.get()
    assert message.body == body


async def test_return_message(amqp_connection: aiormq.Connection):
    conn: aiormq.Connection = amqp_connection
    body = os.urandom(512)
    routing_key = hexlify(os.urandom(16)).decode()

    channel: aiormq.Channel = await conn.channel(
        on_return_raises=False,
    )

    result: Optional[Basic.Return] = await channel.basic_publish(
        body, routing_key=routing_key, mandatory=True,
    )

    assert result is not None

    assert result.delivery.name == "Basic.Return"
    assert result.delivery.routing_key == routing_key


async def test_cancel_on_queue_deleted(amqp_connection):
    conn: aiormq.Connection = amqp_connection
    channel: aiormq.Channel = await conn.channel()
    deaclare_ok = await channel.queue_declare(auto_delete=True)
    consume_ok = await channel.basic_consume(deaclare_ok.queue, print)

    assert consume_ok.consumer_tag in channel.consumers

    with pytest.raises(aiormq.DuplicateConsumerTag):
        await channel.basic_consume(
            deaclare_ok.queue, print, consumer_tag=consume_ok.consumer_tag,
        )

    await channel.queue_delete(deaclare_ok.queue)

    await asyncio.sleep(0.1)

    assert consume_ok.consumer_tag not in channel.consumers


URL_VHOSTS = [
    ("amqp:///", "/"),
    ("amqp:////", "/"),
    ("amqp:///test", "test"),
    ("amqp:////test", "/test"),
    ("amqp://localhost/test", "test"),
    ("amqp://localhost//test", "/test"),
    ("amqps://localhost:5678//test", "/test"),
    ("amqps://localhost:5678/-test", "-test"),
    ("amqp://guest:guest@localhost/@test", "@test"),
    ("amqp://guest:guest@localhost//@test", "/@test"),
]


async def test_ssl_verification_fails_without_trusted_ca(
    amqp_direct_url, rabbitmq_container,
):
    url = amqp_direct_url.with_scheme("amqps").with_port(
        rabbitmq_container.ports["5671/tcp"],
    )
    with pytest.raises(ConnectionError, match=".*CERTIFICATE_VERIFY_FAILED.*"):
        connection = aiormq.Connection(url)
        await connection.connect()


async def test_ssl_context(amqp_direct_url, rabbitmq_container):
    url = amqp_direct_url.with_scheme("amqps").with_port(
        rabbitmq_container.ports["5671/tcp"],
    )
    context = ssl.create_default_context(
        purpose=ssl.Purpose.SERVER_AUTH,
        cafile=cert_path("ca.pem"),
    )
    context.load_cert_chain(cert_path("client.pem"), cert_path("client.key"))
    context.check_hostname = False
    context.verify_mode = ssl.CERT_NONE
    connection = aiormq.Connection(url, context=context)
    await connection.connect()
    await connection.close()


@pytest.mark.parametrize("url,vhost", URL_VHOSTS)
async def test_connection_urls_vhosts(url, vhost, event_loop):
    assert aiormq.Connection(url, loop=event_loop).vhost == vhost


async def test_update_secret(amqp_connection, amqp_url: URL):
    respone = await amqp_connection.update_secret(
        amqp_url.password, timeout=1,
    )
    assert isinstance(respone, aiormq.spec.Connection.UpdateSecretOk)


@aiomisc.timeout(20)
async def test_connection_stuck(proxy, amqp_url: URL):
    url = amqp_url.with_host(
        proxy.proxy_host,
    ).with_port(
        proxy.proxy_port,
    ).update_query(heartbeat="1")

    connection = await aiormq.connect(url)

    async with connection:
        # delay the delivery of each packet by 5 seconds, which
        # is more than the heartbeat
        with proxy.slowdown(50, 50):
            while connection.is_opened:
                await asyncio.sleep(1)

        writer_task: asyncio.Task = connection._writer_task      # type: ignore

        assert writer_task.done()

        with pytest.raises(asyncio.CancelledError):
            assert writer_task.result()

        reader_task: asyncio.Task = connection._reader_task      # type: ignore

        assert reader_task.done()

        with pytest.raises(asyncio.CancelledError):
            assert reader_task.result()


class BadNetwork:
    def __init__(self, proxy, stair: int, disconnect_time: float):
        self.proxy = proxy
        self.stair = stair
        self.disconnect_time = disconnect_time
        self.num_bytes = 0
        self.loop = asyncio.get_event_loop()
        self.lock = asyncio.Lock()

        proxy.set_content_processors(
            self.client_to_server,
            self.server_to_client,
        )

    async def disconnect(self):
        async with self.lock:
            await asyncio.sleep(self.disconnect_time)
            await self.proxy.disconnect_all()
            self.stair *= 2
            self.num_bytes = 0

    async def server_to_client(self, chunk: bytes) -> bytes:
        async with self.lock:
            self.num_bytes += len(chunk)
            if self.num_bytes < self.stair:
                return chunk
            self.loop.create_task(self.disconnect())
            return chunk

    @staticmethod
    def client_to_server(chunk: bytes) -> bytes:
        return chunk


DISCONNECT_OFFSETS = [2 << i for i in range(1, 10)]
STAIR_STEPS = list(
    itertools.product([0.0, 0.005, 0.05, 0.1], DISCONNECT_OFFSETS),
)
STAIR_STEPS_IDS = [
    f"[{i // len(DISCONNECT_OFFSETS)}] {t}-{s}"
    for i, (t, s) in enumerate(STAIR_STEPS)
]


@aiomisc.timeout(30)
@pytest.mark.parametrize(
    "disconnect_time,stair", STAIR_STEPS,
    ids=STAIR_STEPS_IDS,
)
async def test_connection_close_stairway(
    disconnect_time: float, stair: int, proxy, amqp_url: URL,
):
    url = amqp_url.with_host(
        proxy.proxy_host,
    ).with_port(
        proxy.proxy_port,
    ).update_query(heartbeat="1")

    BadNetwork(proxy, stair, disconnect_time)

    async def run():
        connection = await aiormq.connect(url)
        queue = asyncio.Queue()
        channel = await connection.channel()
        declare_ok = await channel.queue_declare(auto_delete=True)
        await channel.basic_consume(
            declare_ok.queue, queue.put, no_ack=True,
        )

        while True:
            await channel.basic_publish(
                b"test", routing_key=declare_ok.queue,
            )
            message: DeliveredMessage = await queue.get()
            assert message.body == b"test"

    for _ in range(5):
        with pytest.raises(aiormq.AMQPError):
            await run()


async def test_ssl_context_provider_static(event_loop):
    certs = SSLCerts(
        cert=None,
        key=None,
        capath=None,
        cafile=None,
        cadata=None,
        verify=False,
    )

    static_context = ssl.create_default_context()
    provider = SSLContextProvider(
        ssl_context=static_context,
        ssl_certs=certs,
        loop=event_loop
    )

    provided_context = await provider.get_context()
    assert provided_context is static_context


async def test_ssl_context_provider_created(event_loop):
    certs = SSLCerts(
        cert=cert_path("client.pem"),
        key=cert_path("client.key"),
        capath=None,
        cafile=cert_path("ca.pem"),
        cadata=None,
        verify=True,
    )

    default_context = ssl.create_default_context()

    provider = SSLContextProvider(
        ssl_context=None,
        ssl_certs=certs,
        loop=event_loop
    )

    provided_context = await provider.get_context()
    assert provided_context != default_context

    second_provided_context = await provider.get_context()
    assert provided_context is second_provided_context


PARSE_INT_PARAMS = (
    (1, 1),
    ("1", 1),
    ("0.1", 0),
    ("-1", -1),
)


@pytest.mark.parametrize("value,expected", PARSE_INT_PARAMS)
def test_parse_int(value, expected):
    assert parse_int(value) == expected


PARSE_TIMEOUT_PARAMS = (
    (1, 1),
    (1.0, 1),
    ("0", 0),
    ("0.0", 0),
    ("0.111", 0.111),
)


@pytest.mark.parametrize("value,expected", PARSE_TIMEOUT_PARAMS)
def test_parse_timeout(value, expected):
    assert parse_timeout(value) == expected


PARSE_BOOL_PARAMS = (
    ("no", False),
    ("nope", False),
    ("do not do it bro", False),
    ("YES", True),
    ("yes", True),
    ("yeS", True),
    ("yEs", True),
    ("True", True),
    ("true", True),
    ("TRUE", True),
    ("1", True),
    ("ENABLE", True),
    ("ENAbled", True),
    ("y", True),
    ("Y", True),
)


@pytest.mark.parametrize("value,expected", PARSE_BOOL_PARAMS)
def test_parse_bool(value, expected):
    assert parse_bool(value) == expected



import asyncio
import atexit
import gc
import logging
import os
import socket
import tracemalloc
from contextlib import suppress
from time import sleep
from typing import Any, Callable, Generator

import pamqp
import pytest
from aiomisc_pytest import TCPProxy
from yarl import URL

from aiormq import Connection

from .docker_client import (
    ContainerInfo,
    DockerClient,
    DockerHostInfo,
    DockerNotAvailableError,
    check_docker_available,
)

# Cached docker host info from pytest_configure
_docker_host_info: DockerHostInfo | None = None

# Global registry for atexit cleanup
_docker_client: DockerClient | None = None
_docker_containers: set[str] = set()


def _atexit_kill_containers() -> None:
    """Kill all containers on exit (handles crashes/interrupts)."""
    if _docker_client is None:
        return
    for container_id in _docker_containers:
        with suppress(Exception):
            _docker_client.kill(container_id)
        with suppress(Exception):
            _docker_client.remove(container_id)
    _docker_containers.clear()


atexit.register(_atexit_kill_containers)


def cert_path(*args):
    return os.path.join(
        os.path.abspath(os.path.dirname(__file__)), "certs", *args,
    )


def pytest_configure(config: pytest.Config) -> None:
    """Check Docker availability before running tests."""
    global _docker_host_info
    if os.environ.get("AMQP_URL"):
        return
    try:
        _docker_host_info = check_docker_available()
    except DockerNotAvailableError as e:
        raise pytest.UsageError(str(e)) from e


@pytest.fixture(scope="session")
def docker() -> Generator[Callable[..., ContainerInfo], Any, Any]:
    global _docker_client
    _docker_client = DockerClient(_docker_host_info)

    def docker_run(
        image: str, ports: list[str],
        environment: dict[str, str] | None = None,
    ) -> ContainerInfo:
        info = _docker_client.run(image, ports, environment=environment)
        _docker_containers.add(info.id)
        return info

    try:
        yield docker_run
    finally:
        for container_id in list(_docker_containers):
            with suppress(Exception):
                _docker_client.kill(container_id)
            with suppress(Exception):
                _docker_client.remove(container_id)
            _docker_containers.discard(container_id)


@pytest.fixture(scope="session")
def rabbitmq_container(docker) -> ContainerInfo:
    amqp_url = os.environ.get("AMQP_URL")
    if amqp_url:
        url = URL(amqp_url)
        return ContainerInfo(
            id="ci-service",
            ports={
                "5672/tcp": url.port or 5672,
                "5671/tcp": 5671,
                "15672/tcp": 15672,
                "15671/tcp": 15671,
            },
            host=url.host or "localhost",
        )
    info = docker(
        "mosquito/aiormq-rabbitmq",
        ["5672/tcp", "5671/tcp", "15672/tcp", "15671/tcp"],
    )
    # Readiness probe - wait for RabbitMQ to be ready
    while True:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            try:
                sock.connect((info.host, info.ports["5672/tcp"]))
                sock.send(b"AMQP\x00\x00\x09\x01")
                data = sock.recv(4)
                if len(data) == 4:
                    return info
            except ConnectionError:
                pass
        sleep(0.3)


@pytest.fixture(scope="session")
def amqp_direct_url(rabbitmq_container: ContainerInfo) -> URL:
    return URL.build(
        scheme="amqp", user="guest", password="guest", path="//",
        host=rabbitmq_container.host,
        port=rabbitmq_container.ports["5672/tcp"],
    )


amqp_url_ids = ["amqp", "amqp-named", "amqps", "amqps-client"]


@pytest.fixture(params=amqp_url_ids, ids=amqp_url_ids)
async def amqp_url(request, amqp_direct_url, rabbitmq_container):
    urls = {
        "amqp": amqp_direct_url,
        "amqp-named": amqp_direct_url.update_query(name="pytest"),
        "amqps": amqp_direct_url.with_scheme("amqps").with_port(
            rabbitmq_container.ports["5671/tcp"],
        ).with_query(
            {"cafile": cert_path("ca.pem"), "no_verify_ssl": 1},
        ),
        "amqps-client": amqp_direct_url.with_scheme("amqps").with_port(
            rabbitmq_container.ports["5671/tcp"],
        ).with_query(
            {
                "cafile": cert_path("ca.pem"),
                "keyfile": cert_path("client.key"),
                "certfile": cert_path("client.pem"),
                "no_verify_ssl": 1,
            },
        ),
    }
    return urls[request.param]


@pytest.fixture
async def amqp_connection(amqp_url, event_loop):
    connection = Connection(amqp_url, loop=event_loop)
    async with connection:
        yield connection


channel_params = [
    dict(channel_number=None, frame_buffer_size=10, publisher_confirms=True),
    dict(channel_number=None, frame_buffer_size=1, publisher_confirms=True),
    dict(channel_number=None, frame_buffer_size=10, publisher_confirms=False),
    dict(channel_number=None, frame_buffer_size=1, publisher_confirms=False),
]


@pytest.fixture(params=channel_params)
async def amqp_channel(request, amqp_connection):
    try:
        yield await amqp_connection.channel(**request.param)
    finally:
        await amqp_connection.close()



@pytest.fixture(autouse=True)
def memory_tracer():
    tracemalloc.start()
    tracemalloc.clear_traces()

    filters = (
        tracemalloc.Filter(True, pamqp.__file__),
        tracemalloc.Filter(True, asyncio.__file__),
    )

    snapshot_before = tracemalloc.take_snapshot().filter_traces(filters)

    def format_stat(stats):
        items = [
            "TOP STATS:",
            "%-90s %6s %6s %6s" % ("Traceback", "line", "size", "count"),
        ]

        for stat in stats:
            fname = stat.traceback[0].filename
            lineno = stat.traceback[0].lineno
            items.append(
                "%-90s %6s %6s %6s"
                % (fname, lineno, stat.size_diff, stat.count_diff),
            )

        return "\n".join(items)

    try:
        yield

        gc.collect()

        snapshot_after = tracemalloc.take_snapshot().filter_traces(filters)

        top_stats = snapshot_after.compare_to(
            snapshot_before, "lineno", cumulative=True,
        )

        if top_stats:
            logging.error(format_stat(top_stats))
            raise AssertionError("Possible memory leak")
    finally:
        tracemalloc.stop()


@pytest.fixture()
async def proxy(tcp_proxy, localhost, amqp_url: URL):
    port = amqp_url.port or 5672 if amqp_url.scheme == "amqp" else 5671
    async with tcp_proxy(amqp_url.host, port) as proxy:
        yield proxy


@pytest.fixture
async def proxy_connection(proxy: TCPProxy, amqp_url: URL, event_loop):
    url = amqp_url.with_host(
        "localhost",
    ).with_port(
        proxy.proxy_port,
    )
    connection = Connection(url, loop=event_loop)

    await connection.connect()

    try:
        yield connection
    finally:
        await connection.close()



import asyncio
import uuid
from os import urandom

import pytest
from aiomisc_pytest import TCPProxy

import aiormq
from aiormq.abc import DeliveredMessage


async def test_simple(amqp_channel: aiormq.Channel):
    await amqp_channel.basic_qos(prefetch_count=1)
    assert amqp_channel.number

    queue = asyncio.Queue()

    deaclare_ok = await amqp_channel.queue_declare(auto_delete=True)
    consume_ok = await amqp_channel.basic_consume(deaclare_ok.queue, queue.put)
    await amqp_channel.basic_publish(
        b"foo",
        routing_key=deaclare_ok.queue,
        properties=aiormq.spec.Basic.Properties(message_id="123"),
    )

    message: DeliveredMessage = await queue.get()
    assert message.body == b"foo"

    cancel_ok = await amqp_channel.basic_cancel(consume_ok.consumer_tag)
    assert cancel_ok.consumer_tag == consume_ok.consumer_tag
    assert cancel_ok.consumer_tag not in amqp_channel.consumers
    await amqp_channel.queue_delete(deaclare_ok.queue)

    deaclare_ok = await amqp_channel.queue_declare(auto_delete=True)
    await amqp_channel.basic_publish(b"foo bar", routing_key=deaclare_ok.queue)

    message = await amqp_channel.basic_get(deaclare_ok.queue, no_ack=True)
    assert message.body == b"foo bar"


async def test_blank_body(amqp_channel: aiormq.Channel):
    await amqp_channel.basic_qos(prefetch_count=1)
    assert amqp_channel.number

    queue = asyncio.Queue()

    deaclare_ok = await amqp_channel.queue_declare(auto_delete=True)
    consume_ok = await amqp_channel.basic_consume(deaclare_ok.queue, queue.put)
    await amqp_channel.basic_publish(
        b"",
        routing_key=deaclare_ok.queue,
        properties=aiormq.spec.Basic.Properties(message_id="123"),
    )

    message: DeliveredMessage = await queue.get()
    assert message.body == b""

    cancel_ok = await amqp_channel.basic_cancel(consume_ok.consumer_tag)
    assert cancel_ok.consumer_tag == consume_ok.consumer_tag
    assert cancel_ok.consumer_tag not in amqp_channel.consumers
    await amqp_channel.queue_delete(deaclare_ok.queue)

    deaclare_ok = await amqp_channel.queue_declare(auto_delete=True)
    await amqp_channel.basic_publish(b"foo bar", routing_key=deaclare_ok.queue)

    message = await amqp_channel.basic_get(deaclare_ok.queue, no_ack=True)
    assert message.body == b"foo bar"


async def test_bad_consumer(amqp_channel: aiormq.Channel, event_loop):
    channel: aiormq.Channel = amqp_channel
    await channel.basic_qos(prefetch_count=1)

    declare_ok = await channel.queue_declare()

    future = event_loop.create_future()

    await channel.basic_publish(b"urgent", routing_key=declare_ok.queue)

    consumer_tag = event_loop.create_future()

    async def bad_consumer(message):
        await channel.basic_cancel(await consumer_tag)
        future.set_result(message)
        raise Exception

    consume_ok = await channel.basic_consume(
        declare_ok.queue, bad_consumer, no_ack=False,
    )

    consumer_tag.set_result(consume_ok.consumer_tag)

    message = await future
    await channel.basic_reject(message.delivery.delivery_tag, requeue=True)
    assert message.body == b"urgent"

    future = event_loop.create_future()

    await channel.basic_consume(
        declare_ok.queue, future.set_result, no_ack=True,
    )

    message = await future

    assert message.body == b"urgent"
    assert message.delivery_tag is not None
    assert message.exchange is not None
    assert message.redelivered


async def test_ack_nack_reject(amqp_channel: aiormq.Channel):
    channel: aiormq.Channel = amqp_channel
    await channel.basic_qos(prefetch_count=1)

    declare_ok = await channel.queue_declare(auto_delete=True)
    queue = asyncio.Queue()

    await channel.basic_consume(declare_ok.queue, queue.put, no_ack=False)

    await channel.basic_publish(b"rejected", routing_key=declare_ok.queue)
    message: DeliveredMessage = await queue.get()
    assert message.body == b"rejected"
    assert message.delivery_tag is not None
    assert message.exchange is not None
    assert not message.redelivered
    await channel.basic_reject(message.delivery.delivery_tag, requeue=False)

    await channel.basic_publish(b"nacked", routing_key=declare_ok.queue)
    message = await queue.get()
    assert message.body == b"nacked"
    await channel.basic_nack(message.delivery.delivery_tag, requeue=False)

    await channel.basic_publish(b"acked", routing_key=declare_ok.queue)
    message = await queue.get()
    assert message.body == b"acked"
    await channel.basic_ack(message.delivery.delivery_tag)


async def test_confirm_multiple(amqp_channel: aiormq.Channel):
    """
    RabbitMQ has been observed to send confirmations in a strange pattern
    when publishing simultaneously where only some messages are delivered
    to a queue. It sends acks like this 1 2 4 5(multiple, confirming also 3).
    This test is probably inconsequential without publisher_confirms
    This is a regression for https://github.com/mosquito/aiormq/issues/10
    """
    channel: aiormq.Channel = amqp_channel
    exchange = uuid.uuid4().hex
    await channel.exchange_declare(exchange, exchange_type="topic")
    try:
        declare_ok = await channel.queue_declare(exclusive=True)
        await channel.queue_bind(
            declare_ok.queue, exchange, routing_key="test.5",
        )

        for i in range(10):
            messages = [
                asyncio.ensure_future(
                    channel.basic_publish(
                        b"test", exchange=exchange,
                        routing_key="test.{}".format(i),
                    ),
                )
                for i in range(10)
            ]
            _, pending = await asyncio.wait(messages, timeout=0.2)
            assert not pending, "not all publishes were completed (confirmed)"
            await asyncio.sleep(0.05)
    finally:
        await channel.exchange_delete(exchange)


async def test_exclusive_queue_locked(amqp_connection):
    channel0 = await amqp_connection.channel()
    channel1 = await amqp_connection.channel()

    qname = str(uuid.uuid4())

    await channel0.queue_declare(qname, exclusive=True)

    try:
        await channel0.basic_consume(qname, print, exclusive=True)

        with pytest.raises(aiormq.exceptions.ChannelLockedResource):
            await channel1.queue_declare(qname)
            await channel1.basic_consume(qname, print, exclusive=True)
    finally:
        await channel0.queue_delete(qname)


async def test_remove_writer_when_closed(amqp_channel: aiormq.Channel):
    with pytest.raises(aiormq.exceptions.ChannelClosed):
        await amqp_channel.queue_declare(
            "amq.forbidden_queue_name", auto_delete=True,
        )

    with pytest.raises(aiormq.exceptions.ChannelInvalidStateError):
        await amqp_channel.queue_delete("amq.forbidden_queue_name")


async def test_proxy_connection(proxy_connection, proxy: TCPProxy):
    channel: aiormq.Channel = await proxy_connection.channel()
    await channel.queue_declare(auto_delete=True)


async def test_declare_queue_timeout(proxy_connection, proxy: TCPProxy):
    for _ in range(3):
        channel: aiormq.Channel = await proxy_connection.channel()

        qname = str(uuid.uuid4())

        with proxy.slowdown(read_delay=5, write_delay=0):
            with pytest.raises(asyncio.TimeoutError):
                await channel.queue_declare(
                    qname, auto_delete=True, timeout=0.5,
                )


async def test_big_message(amqp_channel: aiormq.Channel):
    size = 20 * 1024 * 1024
    message = urandom(size)
    await amqp_channel.basic_publish(message)


async def test_routing_key_too_large(amqp_channel: aiormq.Channel):
    routing_key = "x" * 256

    with pytest.raises(ValueError):
        await amqp_channel.basic_publish(b"foo bar", routing_key=routing_key)

    exchange = uuid.uuid4().hex
    await amqp_channel.exchange_declare(exchange, exchange_type="topic")

    with pytest.raises(ValueError):
        await amqp_channel.exchange_bind(exchange, exchange, routing_key)

    with pytest.raises(ValueError):
        await amqp_channel.exchange_unbind(exchange, exchange, routing_key)

    queue = await amqp_channel.queue_declare(exclusive=True)

    with pytest.raises(ValueError):
        await amqp_channel.queue_bind(queue.queue, exchange, routing_key)

    with pytest.raises(ValueError):
        await amqp_channel.queue_unbind(queue.queue, exchange, routing_key)

    await amqp_channel.exchange_delete(exchange)



import http.client
import json
import os
import socket
from dataclasses import dataclass
from urllib.parse import urlparse


DEFAULT_DOCKER_SOCKET = "/var/run/docker.sock"


class DockerNotAvailableError(Exception):
    """Raised when Docker is not available."""


@dataclass
class DockerHostInfo:
    """Parsed DOCKER_HOST connection info."""
    is_tcp: bool
    # For Unix socket
    socket_path: str | None = None
    # For TCP
    host: str | None = None
    port: int | None = None

    @property
    def container_host(self) -> str:
        """Host where container ports are accessible."""
        if self.is_tcp:
            return self.host or "127.0.0.1"
        return "127.0.0.1"


def parse_docker_host(docker_host: str | None = None) -> DockerHostInfo:
    """Parse DOCKER_HOST environment variable or use default."""
    if docker_host is None:
        docker_host = os.environ.get("DOCKER_HOST", "")

    if not docker_host:
        return DockerHostInfo(
            is_tcp=False,
            socket_path=DEFAULT_DOCKER_SOCKET,
        )

    if docker_host.startswith("unix://"):
        return DockerHostInfo(
            is_tcp=False,
            socket_path=docker_host[7:],
        )

    if docker_host.startswith("tcp://"):
        parsed = urlparse(docker_host)
        return DockerHostInfo(
            is_tcp=True,
            host=parsed.hostname or "127.0.0.1",
            port=parsed.port or 2375,
        )

    # Assume it's a socket path
    return DockerHostInfo(
        is_tcp=False,
        socket_path=docker_host,
    )


def check_docker_available() -> DockerHostInfo:
    """
    Check if Docker is available and raise DockerNotAvailableError if not.

    Returns DockerHostInfo on success for use by DockerClient.
    """
    docker_host_env = os.environ.get("DOCKER_HOST", "")
    info = parse_docker_host(docker_host_env)

    if info.is_tcp:
        # TCP connection to remote Docker
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(5)
            sock.connect((info.host, info.port))
            sock.close()
            return info
        except (socket.error, OSError) as e:
            raise DockerNotAvailableError(
                f"Cannot connect to Docker daemon at "
                f"tcp://{info.host}:{info.port}: {e}\n"
                "\n"
                "Docker is required to run the test suite.\n"
                "\n"
                "Possible solutions:\n"
                "  1. Check if Docker daemon is running on remote host\n"
                "  2. Verify DOCKER_HOST is correct\n"
                "  3. Check firewall/network connectivity\n",
            ) from e
    else:
        # Unix socket connection
        socket_path = info.socket_path
        assert socket_path is not None

        if not os.path.exists(socket_path):
            hints = [
                "Docker is required to run the test suite.",
                "",
                "Possible solutions:",
                "  1. Start Docker daemon",
                "  2. Set DOCKER_HOST environment variable:",
                "     export DOCKER_HOST=unix:///var/run/docker.sock",
                "     export DOCKER_HOST=tcp://remote-host:2375",
                "",
            ]
            if docker_host_env:
                hints.insert(
                    0,
                    f"DOCKER_HOST is set to '{docker_host_env}' "
                    f"but socket '{socket_path}' does not exist.",
                )
            else:
                hints.insert(
                    0,
                    f"Docker socket not found at '{socket_path}' "
                    "and DOCKER_HOST is not set.",
                )
            raise DockerNotAvailableError("\n".join(hints))

        try:
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            sock.settimeout(5)
            sock.connect(socket_path)
            sock.close()
            return info
        except (socket.error, OSError) as e:
            raise DockerNotAvailableError(
                f"Cannot connect to Docker daemon at '{socket_path}': {e}\n"
                "\n"
                "Docker is required to run the test suite.\n"
                "\n"
                "Possible solutions:\n"
                "  1. Start Docker daemon\n"
                "  2. Check Docker socket permissions\n"
                "  3. Add your user to the 'docker' group\n",
            ) from e


@dataclass
class ContainerInfo:
    id: str
    ports: dict[str, int]  # {"5672/tcp": 32789, "5671/tcp": 32790}
    host: str  # "127.0.0.1" for local, remote host for tcp://


class UnixHTTPConnection(http.client.HTTPConnection):
    """HTTP connection over Unix socket for Docker API."""

    def __init__(self, socket_path: str):
        super().__init__("localhost")
        self.socket_path = socket_path

    def connect(self) -> None:
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sock.connect(self.socket_path)


class DockerClient:
    def __init__(self, docker_host_info: DockerHostInfo | None = None):
        if docker_host_info is None:
            docker_host_info = parse_docker_host()
        self.docker_host_info = docker_host_info

    def _get_connection(self) -> http.client.HTTPConnection:
        if self.docker_host_info.is_tcp:
            assert self.docker_host_info.host is not None
            return http.client.HTTPConnection(
                self.docker_host_info.host,
                self.docker_host_info.port,
            )
        assert self.docker_host_info.socket_path is not None
        return UnixHTTPConnection(self.docker_host_info.socket_path)

    def _request(
        self, method: str, path: str, body: dict | None = None,
    ) -> dict | list | None:
        conn = self._get_connection()
        try:
            headers = {"Content-Type": "application/json"} if body else {}
            conn.request(
                method, path,
                body=json.dumps(body) if body else None,
                headers=headers,
            )
            resp = conn.getresponse()
            data = resp.read()
            if resp.status >= 400:
                raise RuntimeError(
                    f"Docker API error: {resp.status} {data.decode()}",
                )
            return json.loads(data) if data else None
        finally:
            conn.close()

    def _request_stream(self, method: str, path: str) -> None:
        """Make a request that returns a stream (e.g., image pull)."""
        conn = self._get_connection()
        try:
            conn.request(method, path)
            resp = conn.getresponse()
            # Consume the stream
            while True:
                chunk = resp.read(4096)
                if not chunk:
                    break
            if resp.status >= 400:
                raise RuntimeError(f"Docker API error: {resp.status}")
        finally:
            conn.close()

    def pull(self, image: str) -> None:
        """Pull an image from registry."""
        # Parse image name and tag
        if ":" in image:
            from_image, tag = image.rsplit(":", 1)
        else:
            from_image, tag = image, "latest"
        path = f"/images/create?fromImage={from_image}&tag={tag}"
        self._request_stream("POST", path)

    def create(
        self, image: str, ports: list[str],
        environment: dict[str, str] | None = None,
    ) -> str:
        """Create a container and return its ID."""
        # Build ExposedPorts config
        exposed_ports: dict[str, dict] = {}
        port_bindings: dict[str, list[dict[str, str]]] = {}
        for port in ports:
            exposed_ports[port] = {}
            # Bind to random host port
            port_bindings[port] = [{"HostIp": "0.0.0.0", "HostPort": ""}]

        # Build environment list
        env_list: list[str] = []
        if environment:
            env_list = [f"{k}={v}" for k, v in environment.items()]

        body = {
            "Image": image,
            "ExposedPorts": exposed_ports,
            "HostConfig": {
                "PortBindings": port_bindings,
                "PublishAllPorts": False,
            },
        }
        if env_list:
            body["Env"] = env_list

        result = self._request("POST", "/containers/create", body)
        assert isinstance(result, dict)
        return result["Id"]

    def start(self, container_id: str) -> None:
        """Start a container."""
        self._request("POST", f"/containers/{container_id}/start")

    def inspect(self, container_id: str) -> dict:
        """Inspect a container."""
        result = self._request("GET", f"/containers/{container_id}/json")
        assert isinstance(result, dict)
        return result

    def stop(self, container_id: str, timeout: int = 10) -> None:
        """Stop a container gracefully."""
        self._request("POST", f"/containers/{container_id}/stop?t={timeout}")

    def kill(self, container_id: str) -> None:
        """Kill a container immediately."""
        self._request("POST", f"/containers/{container_id}/kill")

    def remove(self, container_id: str) -> None:
        """Remove a container."""
        self._request("DELETE", f"/containers/{container_id}")

    def _parse_ports(self, info: dict) -> dict[str, int]:
        """Parse port mappings from container inspect result."""
        ports: dict[str, int] = {}
        network_settings = info.get("NetworkSettings", {})
        port_bindings = network_settings.get("Ports", {})
        for container_port, bindings in port_bindings.items():
            if bindings:
                # Take the first binding's host port
                host_port = int(bindings[0]["HostPort"])
                ports[container_port] = host_port
        return ports

    def run(
        self, image: str, ports: list[str],
        environment: dict[str, str] | None = None,
    ) -> ContainerInfo:
        """Pull, create, start, inspect - return ContainerInfo."""
        self.pull(image)
        container_id = self.create(image, ports, environment=environment)
        self.start(container_id)
        info = self.inspect(container_id)
        return ContainerInfo(
            id=container_id,
            ports=self._parse_ports(info),
            host=self.docker_host_info.container_host,
        )
